# Copyright (c) 2021-2025, ETH Zurich and NVIDIA CORPORATION
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

"""Main module for the rsl_rl package."""

__version__ = "2.1.1"
__license__ = "BSD-3"
